#include <iostream.h>
#include <vector>
#include <ANN/ANN.h>
#include "ann-cc.h"


struct ann_data_t {
  unsigned	_dims;		// what dimension the points are
  unsigned long	_m_pts;		// capacity of `_ptarr' (# of points)
  unsigned long	_n_pts;		// number of points in `_ptarr'; also the
				// first free index in `_ptarr'
  ANNpointArray	_ptarr;
  ANNpoint	_query_pt;
  ANNidxArray	_nn_idx;
  ANNdistArray	_dists;
  ANNkd_tree*	_tree;
};


void*
ann_cc_make_ann(int alloc_num_pts, int dims)
{
  ann_data_t* p = new ann_data_t;
  assert(p);

  if (alloc_num_pts < 1 || dims < 1) {
    cerr << "ann_cc_make_ann(): both parameters must be > 0." << endl;
    return NULL;
  }

  p->_dims = dims;
  p->_m_pts = alloc_num_pts;
  p->_n_pts = 0;
  p->_ptarr = annAllocPts(p->_m_pts, dims);
  p->_query_pt = annAllocPt(dims);
  p->_nn_idx = NULL;		// gets allocated on an actual search
  p->_dists = NULL;		// ditto
  p->_tree = NULL;

  return p;
}


void
ann_cc_del_ann(ann_data_t* p)
{
  if (!p) {
    cerr << "ann_cc_del_ann(): NULL pointer passed!" << endl;
    return;
  }
  
  annDeallocPts(p->_ptarr);
  annDeallocPt(p->_query_pt);
  delete [] p->_nn_idx;
  delete [] p->_dists;
  delete p->_tree;
  delete p;
}


void
ann_cc_add_pt(ann_data_t* p, double* pt, unsigned len)
{
  unsigned i;

  if (!p) {
    cerr << "ann_cc_add_pt(): NULL ann_data passed in!" << endl;
    return;
  }
  
  if (p->_n_pts >= p->_m_pts) {
    // TODO: resize the point array or signal an error to the Python code
    cerr << "ann_cc_add_pt(): point array capacity exceeded." << endl;
    return;
  }

  if (len != p->_dims) {
    cerr << "ann_cc_add_pt(): point is of incorrect length: (expected="
	 << p->_dims << ", actual=" << len << ")" << endl;
    return;
  }

  for (i=0; i<len; i++)
    p->_ptarr[p->_n_pts][i] = pt[i];
  p->_n_pts++;

  // invalidate the tree
  delete p->_tree;
  p->_tree = NULL;
}


void
ann_cc_make_tree(ann_data_t* p)
{
  if (!p) {
    cerr << "ann_cc_make_tree(): NULL ann_data passed in!" << endl;
    return;
  }
  
  if (p->_tree)
    return;

  p->_tree = new ANNkd_tree(p->_ptarr, p->_n_pts, p->_dims);
  assert(p->_tree);
}


// returns number of points actually found
unsigned
ann_cc_search(ann_data_t* p, double* pt, unsigned len, unsigned num_res,
	      double err_bound)
{
  assert(p);

  // prep the query point
  if (len != p->_dims) {
    cerr << "ann_cc_add_pt(): point is of incorrect length: (expected="
	 << p->_dims << ", actual=" << len << ")" << endl;
    return 0;
  }
  for (unsigned i=0; i<len; i++)
    p->_query_pt[i] = pt[i];

  // prep the kd-tree
  ann_cc_make_tree(p);

  // prep the space for the results
  // TODO: don't reallocate these everytime if you can avoid it
  // cleanup
  delete [] p->_nn_idx;
  delete [] p->_dists;
  p->_nn_idx = new ANNidx[num_res];
  p->_dists = new ANNdist[num_res];
  assert(p->_nn_idx && p->_dists);

  // do the search
  if (num_res > p->_n_pts) {
//     cerr << "ann_cc_search(): asking for more neighbours than there are points"
// 	 << endl;
    num_res = p->_n_pts;
  }
  p->_tree->annkSearch(p->_query_pt, num_res, p->_nn_idx, p->_dists, err_bound);

  // count results
  for (unsigned i=0; i<num_res; i++)
    if (p->_nn_idx[i] < 0) {	// negative index means non-solution
      num_res = i;
      break;
    }
      
  // return
  return num_res;
}


// returns last search's solution at index `idx'
unsigned
ann_cc_get_nnidx(ann_data_t* p, unsigned idx)
{
  assert(p);

  // TODO: put in safety check on `idx'
  return p->_nn_idx[idx];
}


// returns last search's solution's distance at `idx'
double
ann_cc_get_dist(ann_data_t* p, unsigned idx)
{
  assert(p);

  // TODO: put in safety check on `idx/
  return p->_dists[idx];
}

